"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import math
import os
import sys

try:
    import ruyaml as yaml
    from ruyaml import constructor, emitter, representer
except ImportError:
    import ruamel.yaml as yaml  # fallback if ruyaml unavailable
    from ruamel.yaml import constructor, emitter, representer  # fallback if ruyaml unavailable

from moat.lib.path import Path

from .dict import attrdict

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    try:
        from ruyaml.constructor import BaseConstructor
        from ruyaml.constructor import SafeConstructor as SafeConstructorType
        from ruyaml.emitter import Emitter as EmitterType
        from ruyaml.nodes import Node
        from ruyaml.representer import BaseRepresenter
        from ruyaml.representer import SafeRepresenter as SafeRepresenterType
    except ImportError:
        from ruamel.yaml.constructor import BaseConstructor  # fallback if ruyaml unavailable
        from ruamel.yaml.constructor import (
            SafeConstructor as SafeConstructorType,  # fallback if ruyaml unavailable
        )
        from ruamel.yaml.emitter import Emitter as EmitterType  # fallback if ruyaml unavailable
        from ruamel.yaml.nodes import Node  # fallback if ruyaml unavailable
        from ruamel.yaml.representer import BaseRepresenter  # fallback if ruyaml unavailable
        from ruamel.yaml.representer import (
            SafeRepresenter as SafeRepresenterType,  # fallback if ruyaml unavailable
        )

    from collections.abc import Callable
    from typing import IO

try:
    from moat.lib.proxy import DProxy, Proxy
except ImportError:
    Proxy = None  # type: ignore[assignment, misc]  # optional import
    DProxy = None  # type: ignore[assignment, misc]  # optional import

__all__ = [
    "add_repr",
    "load_ansible_repr",
    "yaml_parse",
    "yaml_repr",
    "yformat",
    "yload",
    "yprint",
]

SafeRepresenter: type[SafeRepresenterType] = representer.SafeRepresenter
SafeConstructor: type[SafeConstructorType] = constructor.SafeConstructor
Emitter: type[EmitterType] = emitter.Emitter


SafeRepresenter.add_representer(attrdict, SafeRepresenter.represent_dict)


def load_ansible_repr() -> None:
    "Call me if you're using `moat.util` in conjunction with Ansible."

    # optional dependencies, imported at runtime
    from ansible.parsing.yaml.objects import (  # noqa: PLC0415  # type:ignore[unresolved-import]
        AnsibleUnicode,
    )
    from ansible.utils.unsafe_proxy import (  # noqa: PLC0415  # type:ignore[unresolved-import]
        AnsibleUnsafeText,
    )
    from ansible.vars.hostvars import (  # noqa: PLC0415  # type:ignore[unresolved-import]
        HostVars,
        HostVarsVars,
    )

    SafeRepresenter.add_representer(HostVars, SafeRepresenter.represent_dict)
    SafeRepresenter.add_representer(HostVarsVars, SafeRepresenter.represent_dict)
    SafeRepresenter.add_representer(AnsibleUnsafeText, SafeRepresenter.represent_str)
    SafeRepresenter.add_representer(AnsibleUnicode, SafeRepresenter.represent_str)


def str_presenter(dumper: BaseRepresenter, data: str) -> Node:
    """
    Always show multi-line strings with |-style quoting
    """
    if "\n" in data:  # multiline string?
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    else:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def float_presenter(dumper: BaseRepresenter, data: float) -> Node:
    """
    Round appropriately
    """
    if data != 0:
        data = round(data, int(7 - math.log10(abs(data))))
    return dumper.represent_scalar("tag:yaml.org,2002:float", str(data))


def yaml_repr(name: str, use_repr: bool = False) -> Callable[[type], type]:
    """
    A class decorator that allows representing an object in YAML
    """

    def register(cls: type) -> type:
        def str_me(dumper: BaseRepresenter, data: Any) -> Node:
            return dumper.represent_scalar("!" + name, repr(data) if use_repr else str(data))

        SafeRepresenter.add_representer(cls, str_me)
        return cls

    return register


def yaml_parse(name: str, use_repr: bool = False) -> Callable[[type], type]:
    """
    A decorator that allows parsing a YAML representation,
    i.e. the opposite of `yaml_repr`.
    """

    def register(cls: type) -> type:
        SafeConstructor.add_constructor(f"!{name}", cls)
        return cls

    _ = use_repr  # unused
    return register


SafeRepresenter.add_representer(str, str_presenter)
SafeRepresenter.add_representer(float, float_presenter)


def _path_repr(dumper: BaseRepresenter, data: Path) -> Node:
    return dumper.represent_scalar("!P", str(data))


def _proxy_repr(dumper: BaseRepresenter, data: Any) -> Node:
    return dumper.represent_scalar("!R", data.name)


def _dproxy_repr(dumper: BaseRepresenter, data: Any) -> Node:
    return dumper.represent_scalar("!DP", repr([data.name, data.a, data.k]))


def read_env(loader: BaseConstructor, node: Node) -> str:
    value = loader.construct_scalar(node)
    return os.environ[value]


def _el_repr(dumper: BaseRepresenter, data: Any) -> Node:  # noqa: ARG001  # data unused but required by interface
    return dumper.represent_scalar("!R", "...")


SafeRepresenter.add_representer(type(Ellipsis), _el_repr)
SafeRepresenter.add_representer(Path, _path_repr)
SafeConstructor.add_constructor("!P", Path._make)

SafeConstructor.add_constructor("!env", read_env)

if Proxy is not None:
    SafeRepresenter.add_representer(Proxy, _proxy_repr)
if DProxy is not None:
    SafeRepresenter.add_representer(DProxy, _dproxy_repr)


def _name2obj(constructor: BaseConstructor, node: Node) -> Any:
    _ = constructor  # unused but required by interface

    from moat.lib.proxy import name2obj  # noqa: PLC0415

    return name2obj(node.value)


SafeConstructor.add_constructor("!R", _name2obj)


def _bin_from_ascii(loader: BaseConstructor, node: Node) -> bytes:
    value = loader.construct_scalar(node)
    return value.encode("ascii")


def _bin_from_hex(loader: BaseConstructor, node: Node) -> bytearray:
    value = loader.construct_scalar(node)
    return bytearray.fromhex(value.replace(":", ""))


def _bin_to_ascii(dumper: SafeRepresenterType, data: bytes | bytearray | memoryview) -> Node:
    try:
        if isinstance(data, memoryview):
            data_str = data.tobytes().decode("utf-8")
        else:
            data_str = data.decode("utf-8")
    except UnicodeError:
        data_bytes = data.tobytes() if isinstance(data, memoryview) else data
        if len(data_bytes) < 33:
            return dumper.represent_scalar("!hex", data_bytes.hex(":"))
        else:
            return dumper.represent_binary(data_bytes)
    else:
        return dumper.represent_scalar("!bin", data_str)


SafeRepresenter.add_representer(bytes, _bin_to_ascii)
SafeRepresenter.add_representer(bytearray, _bin_to_ascii)
SafeRepresenter.add_representer(memoryview, _bin_to_ascii)

SafeConstructor.add_constructor("!bin", _bin_from_ascii)
SafeConstructor.add_constructor("!hex", _bin_from_hex)


_expect_node = Emitter.expect_node


def expect_node(self: Any, *a: Any, **kw: Any) -> None:
    """
    YAML stream mangler.

    TODO rationale?
    """
    _expect_node(self, *a, **kw)
    self.root_context = False


Emitter.expect_node = expect_node  # type: ignore[method-assign]  # monkey-patch


def yload(
    stream: Any,
    multi: bool = False,
    attr: bool | type[attrdict] = False,
    typ: str = "safe",
) -> Any:
    """
    Load one or more YAML objects from a file.
    """
    y = yaml.YAML(typ=typ)
    if attr:

        class AttrConstructor(SafeConstructor):  # type: ignore[misc, valid-type]  # runtime class creation
            def __init__(self, *a: Any, **k: Any) -> None:
                super().__init__(*a, **k)
                self.yaml_base_dict_type = attrdict if attr is True else attr

        y.Constructor = AttrConstructor
    if multi:
        return y.load_all(stream)
    else:
        return y.load(stream)


def yprint(
    data: Any,
    stream: IO[str] = sys.stdout,
    compact: bool = False,
    typ: str = "safe",
) -> None:
    """
    Write a YAML record.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible, default False.
    """
    if isinstance(data, int | float):
        print(data, file=stream)
    elif isinstance(data, str | bytes):
        print(repr(data), file=stream)
    #   elif isinstance(data, bytes):
    #       os.write(sys.stdout.fileno(), data)
    else:
        y = yaml.YAML(typ=typ)
        y.default_flow_style = compact
        y.dump(data, stream=stream)


def yformat(data: Any, compact: bool | None = None) -> str:
    """
    Return ``data`` as a multi-line YAML string.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible, default False.
    """
    from io import StringIO  # noqa: PLC0415

    s = StringIO()
    yprint(data, compact=compact, stream=s)  # type: ignore[arg-type]  # StringIO is compatible with TextIOWrapper
    return s.getvalue()


def add_repr(typ: type, r: type | None = None) -> None:
    """
    Add a way to add representations for subtypes.

    This is useful for subclassed dict/int/str/… objects.
    """
    if r is None:
        r = typ
    if issubclass(r, str):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_str)
    elif issubclass(r, float):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_float)
    elif issubclass(r, bool):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_bool)
    elif issubclass(r, int):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_int)
    elif issubclass(r, Mapping):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_dict)
    elif issubclass(r, Sequence):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_list)
    else:
        raise TypeError(f"Don't know what to do with {typ}")



"""
This module affords a helper class to make "async with OBJECT" map
seamlessly to an async context management method.
"""

from __future__ import annotations

import anyio
from abc import ABC, abstractmethod
from concurrent.futures import CancelledError  # intentionally not asynio/anyio.Cancelled
from contextlib import AbstractAsyncContextManager, asynccontextmanager

from attrs import define, field

from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from anyio.abc import ObjectReceiveStream, ObjectSendStream, TaskStatus
    from contextvars import ContextVar, Token
    from types import TracebackType

    from collections.abc import AsyncIterator, Awaitable
    from typing import Any, Literal


__all__ = ["CtxObj", "ctx_as", "timed_ctx"]

T_Ctx = TypeVar("T_Ctx")
T_CtxType = TypeVar("T_CtxType")


class CtxObj(ABC, Generic[T_Ctx]):
    """
    Teach a class instance to act as an async context manager, by
    forwarding ``__aenter__`` and ``__aexit__`` to a ``_ctx`` method
    (which must be an ``AsyncContextManager``).

    Usage::

        class Foo(CtxObj):
            @asynccontextmanager
            async def _ctx(self):
                yield self  # or whatever

        async with Foo() as whatever:
            pass
    """

    __ctx: AbstractAsyncContextManager[T_Ctx, bool | None] | None = None

    @abstractmethod
    def _ctx(self) -> AsyncIterator[T_Ctx]: ...

    async def __aenter__(self) -> T_Ctx:
        if self.__ctx is not None:
            raise RuntimeError("Nested contexts")
        ctx_iter = self._ctx()
        if not hasattr(ctx_iter, "__aenter__"):
            # DEPRECATED
            # legacy code for `_ctx` without @asynccm
            ctx: AbstractAsyncContextManager[T_Ctx, bool | None] = asynccontextmanager(self._ctx)()  # type: ignore[assignment]  # legacy support
        else:
            ctx = ctx_iter  # AsyncIterator with __aenter__ is an ACM
        self.__ctx = ctx  # type: ignore[assignment]  # mixed types for legacy support
        return await ctx.__aenter__()  # type: ignore[union-attr]  # both branches have __aenter__

    def __aexit__(
        self,
        *tb: type[BaseException] | None | BaseException | TracebackType,
    ) -> Awaitable[bool | None]:
        try:
            assert self.__ctx is not None
            return self.__ctx.__aexit__(*tb)  # type: ignore[arg-type]  # variadic unpacking
        finally:
            self.__ctx = None


@define
class timed_ctx(CtxObj[T_Ctx]):
    """
    A wrapper for an async context manager that times out if entering it
    takes too long.

    Everything else is unaffected.
    """

    timeout: int | float
    mgr: AbstractAsyncContextManager[T_Ctx, bool | None]

    async def _timer(self, *, task_status: TaskStatus[anyio.CancelScope]) -> None:
        with anyio.CancelScope() as sc:
            task_status.started(sc)
            await anyio.sleep(self.timeout)
            raise TimeoutError(self.timeout)

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[T_Ctx]:
        async with anyio.create_task_group() as tg:
            sc = await tg.start(self._timer)
            async with self.mgr as mgr:
                sc.cancel()
                yield mgr


class ContextMgr(Generic[T_CtxType]):
    """
    This class manages a context for the caller.

    Useful when entering/leaving the context is triggered from state
    machine callbacks or similar event handlers.
    """

    @asynccontextmanager
    async def context(self, *args: Any, **kwargs: Any) -> AsyncIterator[T_CtxType]:  # noqa: ARG002  # abstract method signature
        raise NotImplementedError("Override me!")
        yield  # NotImplementedError prevents reaching this

    exc: Exception | None = None
    ctx: T_CtxType | Literal[False] | None = None
    stopper: anyio.Event | None = None
    stopped: anyio.Event | None = None
    qw: ObjectSendStream[tuple[anyio.Event, tuple[Any, ...], dict[str, Any]]]
    qr: ObjectReceiveStream[tuple[anyio.Event, tuple[Any, ...], dict[str, Any]]]

    def __init__(self) -> None:
        self.qw, self.qr = anyio.create_memory_object_stream(0)

    async def task(self) -> None:
        """
        The task that encapsulates the context handler.

        Start this once, when setting up your state machine.
        """
        async for evt, args, kwargs in self.qr:
            self.stopper = anyio.Event()
            self.stopped = anyio.Event()
            self.exc = None
            try:
                async with self.context(*args, **kwargs) as self.ctx:
                    evt.set()
                    evt = self.stopped  # noqa:PLW2901
                    assert self.stopper is not None  # set above
                    await self.stopper.wait()
                    if self.exc is not None:
                        raise self.exc
            except (CancelledError, Exception) as exc:
                self.exc = exc
            except BaseException:
                self.exc = CancelledError()
                raise
            finally:
                evt.set()
                assert self.stopped is not None  # set above
                self.stopped.set()
                self.ctx = None

    def close(self) -> None:
        """
        Ends the context task.
        """
        self.qw.close()  # type: ignore[attr-defined]  # close() exists on MemoryObjectSendStream
        if self.stopper is not None and not self.stopper.is_set():
            self.exc = CancelledError()
            self.stopper.set()

    async def start(self, *args: Any, **kwargs: Any) -> T_CtxType:
        """
        Creates and starts your context, passing the given arguments.

        Raises `RuntimeError` if the context is already open
        (or starting in a different task).
        """
        if self.ctx is not None:
            raise RuntimeError("Context already entered")
        self.ctx = False
        evt = anyio.Event()
        await self.qw.send((evt, args, kwargs))
        try:
            await evt.wait()
        except BaseException:
            if self.exc is None:
                self.exc = CancelledError()
            assert self.stopper is not None  # set in task()
            self.stopper.set()
            with anyio.move_on_after(0.5, shield=True):
                assert self.stopped is not None  # set in task()
                await self.stopped.wait()
            raise

        if self.exc is not None:
            exc, self.exc = self.exc, None
            raise exc
        assert self.ctx is not False  # set in task() after evt is set
        return self.ctx

    async def stop(self, exc: Exception | None = None) -> None:
        """
        Stops your context.

        If @exc is set, it is passed into / raised in the context.

        This method waits until the context ends.
        """
        if self.ctx is None:
            raise RuntimeError("Context not entered")
        if exc is not None:
            self.exc = exc
        assert self.stopper is not None  # set in task()
        self.stopper.set()
        assert self.stopped is not None  # set in task()
        await self.stopped.wait()
        if self.exc is None:
            return
        if self.exc is exc:
            self.exc = None
        else:
            exc_to_raise, self.exc = self.exc, None
            raise exc_to_raise


@define
class ctx_as:
    """
    Temporary setting of a context variable. This avoids the
    multi-line try/token=set()/finally/reset(token) dance.

    Usage::

        x = ContextVar("x", default=False)
        [async] with ctx_as(x,True):
            assert x.get() is True
        assert x.get() is False  # or whatever
    """

    var: ContextVar[Any] = field()
    value: Any = field()
    token: Token[Any] | None = field(default=None, init=False)

    def __enter__(self) -> Any:
        if self.token is not None:
            raise ValueError("nested 'ctx_as' contexts ??")
        self.token = self.var.set(self.value)
        try:
            return self.value
        finally:
            del self.value

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self.token is not None  # set in __enter__
        self.var.reset(self.token)
        del self.token

    async def __aenter__(self) -> Any:
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.__exit__(exc_type, exc_val, exc_tb)



"""
This module contains various helper functions and classes for dictionaries.
"""

from __future__ import annotations

from copy import deepcopy
from weakref import ref

from moat.lib.path import Path

from . import NotGiven
from ._merge import merge

from collections.abc import Hashable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from types import EllipsisType

    from collections.abc import Callable

__all__ = ["attrdict", "combine_dict", "to_attrdict"]

T = TypeVar("T", bound=dict)

_Nope = object()


def combine_dict(
    *d: Mapping[Hashable, Any] | EllipsisType,
    cls: type[dict] = dict,
    deep: bool = False,
    keep: bool = False,
) -> dict | EllipsisType:
    """
    Returns a dict with all keys+values of all dict arguments.
    The first found value wins.

    This operation is recursive and non-destructive. If ``deep`` is set, the
    result always is a deep copy.

    A value of `NotGiven` causes an entry to be skipped.

    TODO: arrays are not merged.

    Args:
        cls: a class to instantiate the result with. Default: dict.
             Often used: :class:`attrdict`.
        deep: if set, always copy.
        keep: whether to *not* delete `NotGiven` inputs.
    """
    res: dict = cls()
    if not d:
        return res
    if d[0] is NotGiven:
        return NotGiven if keep else res

    keys: dict[Hashable, list[Any]] = {}
    post: bool | None = False if issubclass(cls, attrdict) else None

    if len(d) == 1 and deep and not isinstance(d[0], Mapping):
        if deep and isinstance(d[0], list | tuple):
            return deepcopy(d[0])  # deepcopy returns Any
        else:
            return d[0]  # single non-mapping item

    for kv in d:
        if post is False and getattr(kv, "_post", False):
            post = True
        if kv is None:
            continue
        if kv is NotGiven:
            break
        if not isinstance(kv, dict):
            raise TypeError("All arguments must be dicts")
        for k, v in kv.items():
            if k not in keys:
                keys[k] = []
            keys[k].append(v)

    for k, v in keys.items():
        for i, vv in enumerate(v):
            if isinstance(vv, Path) and vv.is_relative:
                assert isinstance(d[0], attrdict)
                v[i] = d[0].root_.get_(vv)  # d[0] is checked to be attrdict-like
        if v[0] is NotGiven:
            if keep:
                res[k] = NotGiven  # NotGiven is allowed when keep=True
        elif len(v) == 1 and not deep:
            res[k] = v[0]
        elif not isinstance(v[0], Mapping):
            for vv in v[1:]:
                if isinstance(vv, Mapping):
                    raise ValueError(f"Merge {k!r}: {v!r}")  # noqa: TRY004  # ValueError is intentional
            if deep and isinstance(v[0], list | tuple):
                res[k] = deepcopy(v[0])
            else:
                res[k] = v[0]
        else:
            res[k] = combine_dict(*v, cls=cls, keep=keep)

    if post:
        try:
            res.set_post_()  # type: ignore[attr-defined]  # set_post_ exists on attrdict
        except AttributeError:
            pass
    return res


def _check_post(a: Hashable | None, b: Any) -> bool:
    if getattr(b, "_post", False):
        return True
    if isinstance(a, str) and a and a[0] == "$":
        return True
    if isinstance(b, Path):
        return b.is_relative
    if isinstance(b, str):
        return False
    if isinstance(b, Sequence):
        for x in b:
            if _check_post(None, x):  # don't need the index here
                return True
    return False


class attrdict(dict[Hashable, Any]):
    """
    A dictionary which can be accessed via attributes.

    Attributes with leading or trailing underscores are not stored.

    This class supports
    - updating via path accessors
    - recursively tagging updated parents
    """

    _post: bool = False
    _super: ref[attrdict] | None = None

    updated_: Callable[[attrdict], None] = lambda _: None
    "Callback. Used by moat.lib.config."

    def __init__(self, *a: Any, **kw: Any) -> None:
        super().__init__(*a, **kw)
        for a_key, b in self.items():
            if _check_post(a_key, b):
                self._post = True
                return

    @property
    def root_(self) -> attrdict:
        "Follow the _super chain"
        s: attrdict = self
        while True:
            try:
                if (sp := s._super) is None:
                    return s
            except AttributeError:
                return s
            else:
                s_ref = sp()
                if s_ref is None:
                    return s
                s = s_ref

    def __getattr__(self, a: str) -> Any:
        if a.startswith("_") or a.endswith("_"):
            return object.__getattribute__(self, a)
        try:
            return self[a]
        except KeyError:
            raise AttributeError(a) from None

    def __setitem__(self, a: Hashable, b: Any) -> None:
        if isinstance(b, attrdict):
            b._super = ref(self)  # noqa: SLF001  # accessing own class member
        if a not in self or _check_post(a, b) or self[a] != b:
            self._mark_post()
        super().__setitem__(a, b)

    def _mark_post(self) -> None:
        s: attrdict = self
        while True:
            if s._post:
                return
            s._post = True
            sup = s._super
            if not sup:
                return
            sup_ref = sup()
            if not sup_ref:
                return
            s = sup_ref

    def __setattr__(self, a: str, b: Any) -> None:
        if (a and a.startswith("_")) or a.endswith("_"):
            super().__setattr__(a, b)
        else:
            self[a] = b

    def __delattr__(self, a: str) -> None:
        if (a and a.startswith("_")) or a.endswith("_"):
            super().__delattr__(a)
        else:
            try:
                del self[a]
            except KeyError:
                raise AttributeError(a) from None

    def __delitem__(self, a: Hashable) -> None:
        super().__delitem__(a)
        self._mark_post()

    def pop(self, a: Hashable, default: Any = _Nope) -> Any:
        "remove and mark"
        try:
            res = super().pop(a)
        except KeyError:
            if default is _Nope:
                raise
            return default
        self._mark_post()
        return res

    @property
    def needs_post_(self) -> bool:
        """
        Returns a flag whether this attrdict requires postprocessing.

        Warning: Reading this attribute auto-clears the flag.
        """
        try:
            return self._post
        finally:
            self._post = False

    def set_post_(self) -> None:
        """
        Set the flag signalling that this attrdict requires postprocessing.
        """
        self._post = True

    def get_(self, path: Path, default: Any = NotGiven) -> Any:
        """
        Get a node's value and access the dict items beneath it.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")
        val: Any = self
        for p in path:
            if val is None:
                return None
            val = val.get(p, NotGiven)
            if val is NotGiven:
                if default is NotGiven:
                    raise KeyError(path)
                return default
        return val

    _get = get_

    def setdefault(self, a: Hashable, b: Any = None) -> Any:
        """
        Standard dict setdefault but updates _post
        """
        if a in self:
            return self[a]
        if b is None:
            b = None
        self[a] = b
        return b

    def update(  # type: ignore[override]  # More flexible signature than dict.update
        self, a: Mapping[Hashable, Any] | Sequence[tuple[Hashable, Any]] | None = None, **kw: Any
    ) -> None:
        """
        Standard dict update but updates _post
        """
        if a is None:
            pass
        elif isinstance(a, Mapping):
            for k, v in a.items():
                self[k] = v
        else:
            for k, v in a:
                self[k] = v
        for k, v in kw.items():
            self[k] = v

    def update_(self, path: Path, value: Any = None) -> Any:
        """
        Set some sub-item's value, possibly merging dicts.
        Items set to 'NotGiven' are deleted.

        Returns the new value. Modified (sub)dicts will be copied.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")
        val = type(self)()
        val.update(self)
        if getattr(self, "_post", False):
            val.set_post_()

        if not path:
            if isinstance(value, Mapping):
                return combine_dict(value, val, cls=type(self))
            else:
                return value

        px = path[-1]
        post = _check_post(px, value)

        v: Any = val
        if post and isinstance(v, attrdict):
            v.set_post_()
        for p in path[:-1]:
            try:
                w = v[p]
            except KeyError:
                w = type(v)()
            except TypeError:
                if isinstance(v, list) and p is None:
                    p = len(v)  # noqa:PLW2901
                    v.append(None)
                    w = attrdict()
                else:
                    raise
            else:
                # copy
                wx = attrdict() if w is None else type(w)(w)
                if getattr(w, "_post", False):
                    wx.set_post_()
                w = wx
            v[p] = w
            v = w
            if post and isinstance(v, attrdict):
                v.set_post_()

        if value is NotGiven:
            v.pop(px, None)
        elif isinstance(v, list):
            if px is None:
                v.append(value)
                return val
            assert isinstance(px, int)
            if px >= len(v):
                if px > len(v) + 10:
                    raise ValueError(f"Won't pad the array (want {px}, has {len(v)}).")
                v.extend([NotGiven] * (1 + px - len(v)))
            v[px] = value
        elif not isinstance(v, Mapping):
            raise ValueError((v, px))
        elif px in v and isinstance(v[px], Mapping):
            v[px] = value = combine_dict(value, v[px], cls=type(self))
        else:
            v[px] = value

        return val

    _update = update_

    def set_(self, path: Path, value: Any, apply_notgiven: bool = True) -> Any:
        """
        Set some sub-item's value, possibly merging dicts.
        No copying; deleting an entry when the value is NotGiven is optional.

        This function returns Self, *except* when the input path is empty
        and the value is not a Mapping. The latter case is not typed.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")

        if not path:
            if isinstance(value, Mapping):
                return merge(self, value, replace=True)
            else:
                return value

        px = path[-1]
        post = _check_post(px, value)
        v: Any = self

        if post and isinstance(v, attrdict):
            v.set_post_()
        for p in path[:-1]:
            try:
                w = v[p]
            except KeyError:
                w = type(v)()
            except TypeError:
                if isinstance(v, list) and p is None:
                    p = len(v)  # noqa:PLW2901
                    v.append(None)
                    w = attrdict()
                else:
                    raise
            else:
                # create/copy
                if w is None:
                    w = type(v)()
            v[p] = w
            v = w
            if post and isinstance(v, attrdict):
                v.set_post_()

        if value is NotGiven:
            if apply_notgiven:
                v.pop(px, None)
            else:
                v[px] = NotGiven
        elif isinstance(v, list):
            if px is None:
                v.append(value)
                return self
            if px >= len(v):  # type:ignore[unsupported-operator]
                if px > len(v) + 10:  # type:ignore[unsupported-operator]
                    raise ValueError(f"Won't pad the array (want {px}, has {len(v)}).")
                v.extend([NotGiven] * (1 + px - len(v)))  # type:ignore[unsupported-operator]
            v[px] = value
        elif not isinstance(v, Mapping):
            raise ValueError((v, px))
        elif px in v and isinstance(v[px], Mapping):
            merge(v[px], value, replace=True)
        else:
            v[px] = value

        return self

    def delete_(self, path: Path) -> attrdict:
        """
        Remove some sub-item's value, possibly removing now-empty intermediate
        dicts.

        Returns the new value. Modified (sub)dicts will be copied.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")
        val = type(self)(self)
        path_list = list(path)
        v: Any = val
        vc: list[Any] = []
        for p in path_list[:-1]:
            vc.append(v)
            try:
                w = v[p]
            except KeyError:
                return self
            w = type(w)(w)
            v[p] = w
            v = w
        vc.append(v)
        while path_list:
            v = vc.pop()
            v.pop(path_list.pop(), None)
            if v:
                break
        return val

    _delete = delete_


def to_attrdict(d: Any) -> Any:
    """
    Return a hierarchy with all dicts converted to attrdicts.
    """
    if isinstance(d, dict):
        return attrdict((k, to_attrdict(v)) for k, v in d.items())
    if isinstance(d, tuple | list):
        return [to_attrdict(v) for v in d]
    return d



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import sys
from functools import partial
from types import CodeType, ModuleType
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

__all__ = ["Module", "make_module", "make_proc"]


def _call_proc(code: CodeType, variables: Sequence[str], *a: Any, **kw: Any) -> Any:
    v = variables[len(a) :]
    if v:
        a_list = list(a)
        for k in v:
            a_list.append(kw.pop(k, None))
        a = tuple(a_list)
    eval(code, kw)  # noqa: S307
    code_fn = kw["_proc"]
    return code_fn(*a)


def make_proc(
    code: str, variables: Sequence[str], path: Any, *, use_async: bool = False
) -> Callable[..., Any]:
    """Compile this code block to a procedure.

    Args:
        code: the code block to execute. Text, will be indented.
        variables: variable names to pass into the code
        path: the location where the code is stored
        use_async: False if sync code, True if async, None if in thread
    Returns:
        the procedure to call. All keyval arguments will be in the local
        dict.
    """
    hdr = f"""\
def _proc({",".join(variables)}):
    """

    if use_async:
        hdr = "async " + hdr
    code_str = hdr + code.replace("\n", "\n    ")
    code_obj = compile(code_str, str(path), "exec")

    return partial(_call_proc, code_obj, variables)


class Module(ModuleType):
    """A dynamically-loaded module.

    TODO.
    """

    def __repr__(self) -> str:
        return f"<Module {self.__class__.__name__}%s>"


def make_module(code: str, path: Sequence[Any]) -> ModuleType:
    """Compile this code block to something module-ish.

    Args:
        code: the code block to execute
        path: the location where the code is / shall be stored
    Returns:
        the procedure to call. All keyval arguments will be in the local
        dict.
    """
    name = ".".join(str(x) for x in path)
    code_obj = compile(code, name, "exec")
    m = sys.modules.get(name, None)
    if m is None:
        m = ModuleType(name)
    eval(code_obj, m.__dict__)  # noqa: S307
    sys.modules[name] = m
    return m



"""
Exception handling helpers

This code is *duplicated* in moat.micro:_embed/lib/moat/util/exc.py
"""

from __future__ import annotations

import os
from anyio import get_cancelled_exc_class

from moat.lib.micro import log

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Self

__all__ = [
    "ExpAttrError",
    "ExpKeyError",
    "ExpectedError",
    "exc_iter",
    "run_no_exc",
    "ungroup",
]


class ExpectedError(Exception):
    """
    An error that shouldn't elicit a traceback
    """

    def __init__(self, exc: BaseException) -> None:
        self.exc: BaseException = exc


class ExpKeyError(KeyError, ExpectedError):
    "unreported key error"

    pass


class ExpAttrError(AttributeError, ExpectedError):
    "unreported key error"

    pass


def exc_iter(exc: BaseException) -> Iterator[BaseException]:
    """
    iterate over all non-exceptiongroup parts of an exception(group)
    """
    if isinstance(exc, BaseExceptionGroup):
        for e in exc.exceptions:
            yield from exc_iter(e)
    else:
        yield exc


class _Ungroup:
    """
    A sync+async context manager that unwraps single-element
    exception groups.
    """

    def __call__(self) -> Self:
        "Singleton. Returns itself."
        return self

    @staticmethod
    def one(e: BaseException) -> BaseException:
        "convert the exceptiongroup @e to a single exception"
        if not isinstance(e, BaseExceptionGroup):
            return e

        try:
            Cancel = get_cancelled_exc_class()
        except Exception:  # noqa:S110  # no need to be more selective
            pass
        else:
            c, e_split = e.split(Cancel)
            if e_split is not None:
                e = e_split
            elif c is not None:
                e = c

        while isinstance(e, BaseExceptionGroup):
            if len(e.exceptions) != 1:
                break
            e = e.exceptions[0]
        return e

    def __enter__(self) -> Self:
        return self

    async def __aenter__(self) -> Self:
        return self

    def __exit__(
        self,
        c: type[BaseException] | None,
        e: BaseException | None,
        t: object,
    ) -> None:
        if e is None:
            return
        if "MOAT_TB" in os.environ:
            return
        e = self.one(e)
        raise e from None

    async def __aexit__(
        self,
        c: type[BaseException] | None,
        e: BaseException | None,
        t: object,
    ) -> None:
        return self.__exit__(c, e, t)


ungroup: _Ungroup = _Ungroup()


async def run_no_exc(
    p: Callable[..., Any],
    msg: dict[str, Any],
    x_err: tuple[type[Exception], ...] = (),
) -> None:
    """
    Call p(msg) but log exceptions.

    Args:
        p: Callable to execute.
        msg: Keyword arguments to pass to the callable.
        x_err: Exception types to log with reduced detail.
    """
    try:
        r = p(**msg)
        if hasattr(r, "throw"):  # coroutine
            r = await r
    except x_err as err:
        log("Error in %r %r: %r", p, msg, err)
    except Exception as err:
        log("Error in %r %r", p, msg, err=err)



"""Random identifier generation utilities, compatible with µPy."""

from __future__ import annotations

import os

rs = os.environ.get("PYTHONHASHSEED", None)
if rs is None:
    import random
else:  # pragma: no cover
    try:
        import trio._core._run as tcr
    except ImportError:
        import random
    else:
        random = tcr._r  # noqa: SLF001

al_unique: str = "bcdfghjkmnpqrstvwxyzBCDFGHJKMNPQRSTVWXYZ23456789"
"An alphabet intended to be unambiguous: both cases, no special characters, no vowels"

al_lower: str = "abcdefghijklmnopqrstuvwxyz0123456789"
"Lowercase letters and digits, e.g. for restricted-alphabet labels"

al_ascii: str = bytes(x for x in range(33, 127) if x != 92).decode("ascii")
"all printable ASCII characters (except for backslash, just to be safe)"

al_az: str = "abcdefghijklmnopqrstuvwxyz"
"lowercase letters"

__all__ = ["al_ascii", "al_az", "al_lower", "al_unique", "gen_ident", "id2str"]


def gen_ident(k: int = 10, /, *, alphabet: str = al_unique) -> str:
    """
    Generate a random identifier / password.
    """
    return "".join(random.choices(alphabet, k=k))


def id2str(k: int, /, *, alphabet: str = al_unique) -> str:
    """
    Given a non-numeric identifier, returns a string representing it.
    """
    assert k > 0
    if k == 0:
        return alphabet[0]
    res: list[str] = []
    la = len(alphabet)
    while k:
        res.append(alphabet[k % la])
        k //= la
    return "".join(res)



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import anyio

from typing import TYPE_CHECKING, ParamSpec

if TYPE_CHECKING:
    from anyio.abc import TaskGroup, TaskStatus

    from collections.abc import Awaitable, Callable

__all__ = ["spawn"]

P = ParamSpec("P")


async def spawn(
    taskgroup: TaskGroup,
    proc: Callable[P, Awaitable[object]],
    *args: P.args,
    **kw: P.kwargs,
) -> anyio.CancelScope:
    """
    Run a task within this object's task group.

    Returns:
        a cancel scope you can use to stop the task.
    """

    async def _run(
        *,
        task_status: TaskStatus[anyio.CancelScope],
    ) -> None:
        """
        Helper for starting a task within a cancel scope.
        """
        with anyio.CancelScope() as sc:
            task_status.started(sc)
            await proc(*args, **kw)

    return await taskgroup.start(_run)



"""
This module contains a heap of somewhat-random helper functions
and classes which are used throughout MoaT (and beyond)
but don't get their own package because they're too small,
or too interrelated … or the author was too lazy.
"""
# TODO split this up

from __future__ import annotations

import logging as _logging

from typing import Any

_log = _logging.getLogger(__name__)

NotGiven = Ellipsis  # Sentinel value for "no value given"

# Mapping of exported names to their source modules
_imports = {
    # impl
    "Cache": "impl",
    "NoLock": "impl",
    "OptCtx": "impl",
    "SigCancel": "sig",
    "TimeOnlyFormatter": "impl",
    "acount": "impl",
    "byte2num": "impl",
    "count": "impl",
    "digits": "impl",
    "import_": "impl",
    "load_from_cfg": "impl",
    "num2byte": "impl",
    "num2id": "impl",
    "singleton": "impl",
    "split_arg": "impl",
    # dict
    "combine_dict": "dict",
    "drop_dict": "dict",
    "to_attrdict": "dict",
    "attrdict": "dict",
    # merge
    "merge": "_merge",
    # misc
    "OutOfData": "misc",
    "_add_obj": "misc",
    "get_codec": "misc",
    "pos2val": "misc",
    "srepr": "misc",
    "val2pos": "misc",
    "wait_complain": "misc",
    "de_async": "misc",
    # pp
    "pop_kw": "pp",
    "push_kw": "pp",
    # random
    "al_ascii": "random",
    "al_az": "random",
    "al_lower": "random",
    "al_unique": "random",
    "gen_ident": "random",
    "id2str": "random",
    # inexact
    "InexactFloat": "inexact",
    # event
    "ValueEvent": "event",
    # ctx
    "CtxObj": "ctx",
    "ctx_as": "ctx",
    "timed_ctx": "ctx",
    # queue
    "DelayedRead": "queue",
    "DelayedWrite": "queue",
    "Lockstep": "queue",
    "Queue": "queue",
    "QueueEmpty": "queue",
    "QueueFull": "queue",
    "create_queue": "queue",
    # module
    "Module": "module",
    "make_module": "module",
    "make_proc": "module",
    # msg
    "MsgReader": "msg",
    "MsgWriter": "msg",
    # server
    "gen_ssl": "server",
    "run_tcp_server": "server",
    # spawn
    "spawn": "spawn",
    # systemd
    "as_service": "systemd",
    "RunMsg": "systemd",
    # yaml
    "add_repr": "yaml",
    "load_ansible_repr": "yaml",
    "yaml_parse": "yaml",
    "yaml_repr": "yaml",
    "yformat": "yaml",
    "yload": "yaml",
    "yprint": "yaml",
    # exc
    "ExpAttrError": "exc",
    "ExpKeyError": "exc",
    "ExpectedError": "exc",
    "exc_iter": "exc",
    "run_no_exc": "exc",
    "ungroup": "exc",
    # part
    "enc_part": "part",
    "get_part": "part",
    "set_part": "part",
}

__all__ = list(_imports.keys())


def __getattr__(attr: str) -> Any:
    try:
        mod = _imports[attr]
    except KeyError:
        raise AttributeError(attr) from None
    value = getattr(__import__(mod, globals(), None, (), 1), attr)
    globals()[attr] = value
    return value


def __dir__() -> list[str]:
    return __all__



# Trio step monitors

from __future__ import annotations

import sys

from trio._core._instrumentation import Instruments
from trio.abc import Instrument

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from types import FrameType

    from outcome import Outcome
    from trio._core._run import Runner, Task


class Instr(Instrument):
    exited: None = None

    def before_task_step(self, task: Task) -> None:
        print("BEF:", task, task._next_send, file=sys.stderr)  # noqa: SLF001

    def after_task_step(self, task: Task) -> None:
        print("AFT:", task, task._next_send, file=sys.stderr)  # noqa: SLF001

    def task_exited(self, task: Task, outcome: Outcome[Any] | None = None) -> None:
        print("EXIT", task, outcome, file=sys.stderr)


def hookup(runner: Runner | None = None) -> None:
    if runner is None:
        import inspect  # noqa: PLC0415

        f: FrameType | None = inspect.currentframe().f_back  # type: ignore[union-attr]  # currentframe can return None but we know we're in a function
        while f is not None and "runner" not in f.f_locals:
            f = f.f_back
            if f is None:
                return  # no Trio
        runner = f.f_locals["runner"]  # type: ignore[union-attr]  # f is not None here

    i = Instr()
    ig = Instruments([i])
    runner.instruments = ig



"""
This module contains a helper for running subprocesses.
"""

from __future__ import annotations

import anyio
import io
from codecs import getincrementaldecoder
from pathlib import Path
from subprocess import DEVNULL, PIPE, STDOUT, CalledProcessError

from typing import TYPE_CHECKING, cast, overload

if TYPE_CHECKING:
    from io import BytesIO

    from collections.abc import AsyncIterable
    from typing import Literal


__all__ = ["DEVNULL", "PIPE", "STDOUT", "CalledProcessError", "run"]


@overload
async def run(
    *a,
    name: str | None = None,
    echo: bool = False,
    echo_input: bool = False,
    capture=False,
    input: str | bytes | None = None,
    **kw,
) -> None:
    "no data"


@overload
async def run(
    *a,
    name: str | None = None,
    echo: bool = False,
    echo_input: bool = False,
    capture="raw",
    input: str | bytes | None = None,
    **kw,
) -> bytes:
    "raw data"


@overload
async def run(
    *a,
    name: str | None = None,
    echo: bool = False,
    echo_input: bool = False,
    capture=True,
    input: str | bytes | None = None,
    **kw,
) -> str:
    "string data"


async def run(
    *a,
    name=None,
    echo=False,
    echo_input=False,
    capture: bool | Literal["raw"] = False,
    input=None,  # noqa: A002
    **kw,
) -> None | str | bytes:
    """Helper to run an external program, tagging stdout/stderr"""

    if name is None:
        name = ""
    else:
        name += " "
    if isinstance((cwd := kw.get("cwd")), (anyio.Path, Path)):
        kw["cwd"] = str(cwd)
    if echo:
        if cwd is not None:
            print(name + "$", "cd", cwd)
        print(name + "$", *a, *(("<", repr(kw["input"])) if echo_input and "input" in kw else ()))
    if input is None:
        if "stdin" not in kw:
            kw["stdin"] = DEVNULL
    else:
        if isinstance(input, str):
            input = input.encode("utf-8")  # noqa: A001

    if capture and kw.get("stdout", PIPE) != PIPE:
        raise ValueError("can't capture if stdout is not PIPE")

    frag = None
    out = None
    err = None

    async def report(prefix: str, stream: AsyncIterable[bytes], buf: BytesIO):
        nonlocal frag
        utf = getincrementaldecoder("utf-8")()
        async for chunk in stream:
            buf.write(chunk)
            if not echo:
                continue
            chunk = utf.decode(chunk).split("\n")  # noqa:PLW2901
            lch = chunk.pop()
            if frag not in (None, stream):
                print("…")

            for ch in chunk:
                print(prefix, ch)

            if lch:
                print(prefix, lch, end="")
                frag = stream
            else:
                frag = None

    async with (
        await anyio.open_process(a, **kw) as proc,
        anyio.create_task_group() as tg,
    ):
        if proc.stdout:
            out = io.BytesIO()
            tg.start_soon(report, name + ">", proc.stdout, out)
        if proc.stderr:
            err = io.BytesIO()
            tg.start_soon(report, name + "⫸", proc.stderr, err)
        if input is not None:
            assert proc.stdin is not None  # stdin is PIPE when input is provided
            await proc.stdin.send(input)
            await proc.stdin.aclose()

    if frag is not None:
        print("…")

    if proc.returncode != 0:
        raise CalledProcessError(
            cast(int, proc.returncode),
            a,
            None if out is None else out.getvalue(),
            None if err is None else err.getvalue(),
        )

    if capture:
        assert out is not None  # stdout is PIPE when capture is True
        res = out.getvalue()
        if capture is True:
            res = res.decode("utf-8")
        return res

    return None



"""
Floats that compare with a delta.
"""

from __future__ import annotations

from math import isclose

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self

__all__ = ["InexactFloat"]


class InexactFloat(float):
    """This is a wrapper for `float`that supports inexact comparison and
    output of float numbers.

    Comparison is mediated by :py:func:`math.isclose`.

    If both sides are an :py:class:`InexactFloat`, numbers are considered
    equal if either is within the tolerances of the other.

    The number of significant digits does not affect the comparison.

    This class only affects comparison, but not any other mathematical
    operation.

    .. note::
        Inexact equality is not transitive: a=b and b=c is not sufficient
        for a=c.
    """

    __slots__ = ("_abs", "_digits", "_rel")

    _rel: float
    _abs: float
    _digits: int

    def __new__(
        cls,
        val: float,
        rel: float = 1e-06,
        abs: float = 1e-12,  # noqa: A002
        digits: int = 3,
    ) -> Self:
        """
        Args:
            val: the exact value.
            rel: Acceptable relative error.
            abs: Acceptable absolute error.
            digits: Number of significant digits.

        """
        res = super().__new__(cls, val)
        res._rel = rel
        res._abs = abs
        res._digits = digits
        return res

    def __hash__(self) -> int:
        raise TypeError(f"unhashable type: {type(self)}")

    def __eq__(self, b: object) -> bool:
        """Check if the two floats are mostly-equal."""
        if isinstance(b, int | float | InexactFloat) and isclose(
            self, b, rel_tol=self._rel, abs_tol=self._abs
        ):
            return True
        if not isinstance(b, InexactFloat):
            return False
        return isclose(b, self, rel_tol=b._rel, abs_tol=b._abs)

    def __ne__(self, b: object) -> bool:
        return not self.__eq__(b)

    def __lt__(self, other: float | InexactFloat) -> bool:
        return super().__lt__(other) and self != other

    def __gt__(self, other: float | InexactFloat) -> bool:
        return super().__gt__(other) and self != other

    def __le__(self, other: float | InexactFloat) -> bool:
        return super().__le__(other) or self == other

    def __ge__(self, other: float | InexactFloat) -> bool:
        return super().__ge__(other) or self == other

    # --repr
    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __str__(self) -> str:
        return f"{round(self, self._digits):.{self._digits}f}"



"""Signal handling utilities."""

from __future__ import annotations

import anyio
import signal
from contextlib import asynccontextmanager

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from anyio.abc import TaskGroup

    from collections.abc import AsyncIterator


class SigCancel(anyio.AsyncContextManagerMixin):
    """
    This is a cancellation-by-signal helper.

    Usage::

        async with SigCancel():
            ... do whatever

    Args:
        exc:
            Exception to raise on signal. The exception will *not* be wrapped in an
            `ExceptionGroup`. If unset, terminate silently.
    """

    def __init__(self, exc: type[BaseException] | None = None) -> None:
        self.exc: type[BaseException] | None = exc
        self.tg: TaskGroup

    async def _sig_handler(self) -> None:
        with anyio.open_signal_receiver(
            signal.SIGINT,
            signal.SIGTERM,
            signal.SIGHUP,
        ) as signals:
            async for _ in signals:
                self.tg.cancel_scope.cancel()
                break  # default handlers on next

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncIterator[None]:
        async with anyio.create_task_group() as self.tg:
            self.tg.start_soon(self._sig_handler, name="sig")
            yield None
            if self.tg.cancel_scope.cancel_called and self.exc is not None:
                raise self.exc
            self.tg.cancel_scope.cancel()



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import logging
import sys
from contextlib import nullcontext, suppress
from getpass import getpass
from math import log10

from collections import deque
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from types import TracebackType

    from collections.abc import AsyncIterator, Callable, Iterator

__all__ = [
    "Cache",
    "NoLock",
    "NoneType",
    "OptCtx",
    "TimeOnlyFormatter",
    "acount",
    "byte2num",
    "count",
    "digits",
    "import_",
    "load_from_cfg",
    "num2byte",
    "num2id",
    "singleton",
    "split_arg",
]

NoneType = type(None)

NoLock = nullcontext()

T = TypeVar("T")


class OptCtx:
    """
    Optional context. Unlike `contextlib.nullcontext` this doesn't return a
    fixed value; instead it delegates to the wrapped context manager – if
    there is one.
    """

    def __init__(self, obj: object = None) -> None:
        self.obj: object = obj

    def __enter__(self) -> object:
        if self.obj is not None:
            return self.obj.__enter__()  # type: ignore[union-attr]  # duck typing
        return None

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> object:
        if self.obj is not None:
            return self.obj.__exit__(exc_type, exc_val, exc_tb)  # type: ignore[union-attr]  # duck typing
        return None

    async def __aenter__(self) -> object:
        if self.obj is not None:
            return await self.obj.__aenter__()  # type: ignore[union-attr]  # duck typing
        return None

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> object:
        if self.obj is not None:
            return await self.obj.__aexit__(exc_type, exc_val, exc_tb)  # type: ignore[union-attr]  # duck typing
        return None


def import_(name: str, off: int = 0) -> object:
    """
    Import a module and access an object in it.

    ``import_("a.b.c.d.e", 2)`` imports "a.b.c" and returns the e attribute
    of object d from it.
    """
    n = name.split(".")
    mn = ".".join(n[: -off if off else 99])
    try:
        res: object = __import__(mn)
        for nn in n[1:]:
            res = getattr(res, nn)
    except Exception:
        sys.modules.pop(mn, None)
        raise
    return res


def load_from_cfg(
    *a: Any,
    _cfg: dict[str, Any] | None = None,
    _attr: str = "server",
    _raise: bool = False,
    **k: Any,
) -> object | None:
    """
    A simple frontend to load a module, access a class/object from it,
    and call that with the config (and whchever other arguments you want to
    use).

    The module+object name is the "server" attribute (or @_attr).

    Args:
        _cfg: The configuration stanza to use. If not given, a ``cfg``
              argument must exist.
        _attr: The config entry that names the service to load.
               The default is ``"server"``.
        _raise: If no ``_attr`` exists, raise KeyError instead of returning `None`.

    """
    if _cfg is None:
        cfg = k["cfg"]
    else:
        cfg = _cfg
    try:
        name = cfg[_attr]
    except KeyError:
        if _raise:
            raise
        return None
    if isinstance(name, list | tuple):
        name, off = name
    else:
        off = 1
    m = import_(name, off=off)
    return m(*a, **k)  # type: ignore[operator]  # m is callable


def singleton(cls: Callable[[], T]) -> T:
    """Basic singleton decorator"""
    return cls()


class TimeOnlyFormatter(logging.Formatter):
    """A log formatter that doesn't show dates"""

    default_time_format: str = "%H:%M:%S"
    default_msec_format: str = "%s.%03d"


def count(it: Iterator[object]) -> int:
    """counts the length of an iterator"""
    n = 0
    for _ in it:
        n += 1
    return n


async def acount(it: AsyncIterator[object]) -> int:
    """counts the length of an async iterator"""
    n = 0
    async for _ in it:
        n += 1
    return n


class Cache:
    """
    A quick-and-dirty cache that keeps the last N entries of anything
    in memory so that ref and WeakValueDictionary don't lose them.

    Entries get refreshed when they're in the last third of the cache; as
    they're not removed, the actual cache size might only be 2/3rd of SIZE.
    """

    def __init__(self, size: int) -> None:
        self._size: int = size
        self._head: int = 0
        self._tail: int = 0
        self._attr: str = "_cache__pos"
        self._q: deque[object] = deque()

    def keep(self, entry: object) -> None:
        """Store an entry in the cache"""
        if getattr(entry, self._attr, -1) > self._tail + self._size / 3:
            return
        self._head += 1
        setattr(entry, self._attr, self._head)
        self._q.append(entry)
        self._flush()

    def _flush(self) -> None:
        while self._head - self._tail > self._size:
            self._q.popleft()
            self._tail += 1

    def resize(self, size: int) -> None:
        """Change the size of this cache."""
        self._size = size
        self._flush()

    def clear(self) -> None:
        """Clear the cache"""
        while self._head > self._tail:
            self._q.popleft()
            self._tail += 1


def digits(n: float, digits: float = 6) -> float:
    """
    Returns ``n`` rounded to ``digits`` significant digits. Default: 6.
    Ensures that the number doesn't carry nonsense precision or
    floating-point artefacts.

    >>> digits(123456789, 4)
    123400000
    >>> digits(math.pi, 4)
    3.142

    ``digits`` may be a fraction, in order to move the cut-off point to
    somewhere other than between 9.999 and 10.00.
    """
    return round(n, int(digits - 1 - log10(abs(n))))


def num2byte(num: int, length: int | None = None) -> bytes:
    """
    convert an unsigned integer to a bytestring
    """
    if length is None:
        length = (num.bit_length() + 7) // 8
    return num.to_bytes(length=length, byteorder="big")


def byte2num(data: bytes) -> int:
    """
    convert a bytestring to an unsigned integer
    """
    return int.from_bytes(data, byteorder="big")


def split_arg(p: str, kw: dict[str, Any]) -> None:
    """
    Split argument 'p' and add to dict 'kw'.

    'p' may be of the form x=y (y is added, possibly as an integer),
    x? (call getpass(x? )), x?=y (call getpass(y: )).
    """
    k: str
    v: str | int
    try:
        k, v = p.split("=", 1)
    except ValueError:
        if p[-1] == "?":
            k = p[:-1]
            v = getpass(k + "? ")
        else:
            raise
    else:
        if k[-1] == "?":
            k = k[:-1]
            v = getpass(v + ": ")
        with suppress(ValueError):
            v = int(v)
    kw[k] = v


_alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"


def num2id(number: int | object, alphabet: str = _alphabet) -> str:
    """
    Encode a number / object ID in base36 (by default).

    This code doesn't care about num2id(739172) or similar.

    To avoid these issues, pass an alphabet without vowels
    as the second parameter, e.g. :py:data:`moat.util.al_unique`.
    """
    if not isinstance(number, int):
        if isinstance(number, float | complex | str | bytes | bytearray):
            raise TypeError("number must be an object or integer")
        number = id(number)
    is_negative = number < 0
    number = abs(number)
    res = []

    while number:
        number, i = divmod(number, len(alphabet))
        res.append(alphabet[i])
    if is_negative:
        res.append("-")
    elif not res:
        return alphabet[0]
    res.reverse()

    return "".join(res)



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import anyio
from anyio.abc import SocketAttribute

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from ssl import SSLContext

    from anyio.abc import TaskGroup

__all__ = ["gen_ssl", "run_tcp_server"]


class _Server:
    """
    Wrapper for running a server.

    :meta public:
    """

    server: Any = None

    def __init__(
        self,
        tg: TaskGroup,
        handler: Callable[[Any], Awaitable[None]],
        _rdy: Callable[[Any], None] | None = None,
        port: int = 0,
        ssl: SSLContext | None = None,
        **kw: Any,
    ) -> None:
        self.tg: TaskGroup = tg
        self._kw: dict[str, Any] = kw
        self.ssl: SSLContext | None = ssl
        self._rdy: Callable[[Any], None] | None = _rdy
        self.handler: Callable[[Any], Awaitable[None]] = handler
        self.port: int = port

    async def _accept(self, conn: Any) -> None:
        conn_: Any = conn
        if self.ssl:
            conn_ = await anyio.streams.tls.TLSStream.wrap(  # type: ignore[attr-defined]  # anyio.streams.tls available at runtime
                conn,
                server_side=True,
                ssl_context=self.ssl,
            )
        await self.handler(conn_)

    async def run(self) -> None:
        """Run the server."""
        try:
            listener = await anyio.create_tcp_listener(local_port=self.port)
        except Exception as exc:
            raise OSError(f"could not listen to port {self.port}") from exc
        if not self.port:
            self.port = listener.extra_attributes[SocketAttribute.local_address]()[1]
        if self._rdy is not None:
            self._rdy(listener)
        async with listener:
            await listener.serve(self._accept)


async def run_tcp_server(*a: Any, **kv: Any) -> _Server:
    """
    A simple, possibly-SSL TCP server that runs a handler for each incoming connection
    """
    tg = kv.pop("tg", None)
    server: _Server
    if tg is not None:
        server = _Server(tg, *a, **kv)
        await server.run()
    else:
        async with anyio.create_task_group() as tg:
            server = _Server(tg, *a, **kv)
            await server.run()
    return server


def gen_ssl(
    ctx: bool | SSLContext | dict[str, str] = False,
    server: bool = True,
) -> SSLContext | None:
    """
    Generate a SSL config from the given context.

    Args:
      ctx: either a Bool (ssl yes/no) or a dict with "key" and "cert" entries.
      server: a flag whether to behave as a server.
    """
    if not ctx:
        return None
    if ctx is True:
        ctx = {}
    if not isinstance(ctx, dict):
        return ctx

    import ssl  # noqa: PLC0415

    ctx_ = ssl.create_default_context(
        purpose=ssl.Purpose.CLIENT_AUTH if server else ssl.Purpose.SERVER_AUTH,
    )
    if "key" in ctx:
        ctx_.load_cert_chain(ctx["cert"], ctx["key"])
    return ctx_
